#include "iostream.h"

#include <stdio.h>
#include <String.h>
#include <stdlib.h>

FILE *infile, *outfile;  // The 3DS file we will be reading
unsigned long fileSize;

unsigned char	ReadChar(void);
unsigned int	ReadInt(void);
unsigned long	ReadLong(void);
int				ReadLongName(void);
int				ReadTextureFile(void);
float 			ReadFloat(void);

int				Translate(void);
double 			GetLength( unsigned int);
bool 			SetLength(unsigned int, double);

BString			TextureMapName;
BString 		tempName;

struct Point3D {
	float X;
	float Y;
	float Z;
};

Point3D vertex[30000];

struct Triangle {
	unsigned int A;
	unsigned int B;
	unsigned int C;
};

Triangle face[64000];

struct Vector {
	double X;
	double Y;
	double Z;
};

Vector tempVect1;
Vector tempVect2;

Vector vertexVect[30000];

Vector faceVect[64000];
Vector mixFace[120];
Vector tempVect;

unsigned int refs = 0;




double vctrX[500], vctrY[500], vctrZ[500] ;

//  Translation function
int	Translate(void)
{
	unsigned int chunk_ID;
	unsigned long size;
	
	unsigned long current_pointer = 0;

	unsigned long 	TempLong;
	float			TempFloat;

	unsigned int  	faceCount;

	unsigned int	vertexCount;



	
	bool 			end_found	= FALSE;

while (end_found==FALSE)
{
	
	current_pointer = (ftell (infile));
	
	//Output current position for debugging purposes
	//cout << "\n" << current_pointer;   

	if (current_pointer >= fileSize)
	{
		//cout << "\nSuccessfully reached End of File.\n";
		end_found=TRUE;
		return(0); 
	}	

	
	//Read Chunk header information
	chunk_ID = ReadInt();
	size = ReadLong();

	switch (chunk_ID)
	{
	
	//Main Chunk (M3DMAGIC)
		case 0x4D4D:
			cout << "\nMain Chunk - 0x4D4D: " << size << " bytes\n";

			//verify that file is the size reported by Main Chunk
			if (fileSize != size)
			{
				cout << "\n*** Warning: Size of file does not match Main Chunk\n";
				cout <<   "*** Warning: Possible File Corruption\n";
				return(1);  //To be on the safe side....
			}
		break;
		
	//Version Chunk (M3D_VERSION)
		case 0x0002:
			cout << "\n\t3DS Version - 0x0002: "  << size << " bytes - ";

			TempLong = ReadLong();
			cout << "Version: " << TempLong << ".0\n";

			//verify that file is a readable version
			if (TempLong < 3)
			{
				cout << "\n*** Warning: This translator can only read 3DS files of version 3.0 and higher\n" ;
				cout <<   "*** Warning: The version of the file you want to read is: " << TempLong << "\n";
				return(1);  //To be on the safe side....
			}
		break;

	//3D Editor Chunk - MDATA
		case 0x3D3D:
			cout << "\n\t3D Editor Chunk - 0x3D3D: "<< size << " bytes\n";		
		break;

	//Mesh Version Chunk
		case 0x3D3E:
			cout << "\n\t\tMesh Version - 0x3D3E: " << size << " bytes - ";
			TempLong = ReadLong();
			cout << "Version: " << TempLong << ".0\n";
		break;

	//Master_Scale Chunk
		case 0x0100:
			cout << "\n\t\tMaster Scale - 0x0100: "<< size << " bytes - ";
			TempFloat = ReadFloat();
			cout << TempFloat << " units/inch\n";
		break;



	//Named_Object
		case 0x4000:
			ReadLongName();
			cout << "\n\t" << tempName.String() << " - Named Object - 0x4000: " << size << " bytes\n";
		break;

	//Triangular Mesh
		case 0x4100:
			cout << "\n\t\tTriangular Mesh - 0x4100: "<< size << " bytes\n";
		break;

	//Vertex_Array
		case 0x4110:
			vertexCount = ReadInt();
			cout << "\n\t\t\tVertex Array - 0x4110: " << size << " bytes -  ";
			cout << vertexCount <<" entries\n\n";
 
			fprintf(outfile,"%d\n",vertexCount);

			// Read the array into memory
			for (unsigned int i=0; i < vertexCount; i++)
 			{
 				vertex[i].X = ReadFloat();
 				vertex[i].Y = ReadFloat();
 				vertex[i].Z = ReadFloat();
			}

		break;

	//Face_Array
		case 0x4120:

			unsigned int temp_diff;
			faceCount = ReadInt();


			cout << "\n\t\t\tFace Array - 0x4120: "<< size << " bytes -  ";
			cout << faceCount << " entries\n\n";

			// Read the vertex indexes of each face into memory 
			for (unsigned int i = 0; i < faceCount; i++)
			{
				face[i].A = ReadInt();
				face[i].B = ReadInt();
				face[i].C = ReadInt();
				
				//Edge Visibility and UV Wrapping info bits
				temp_diff = ReadInt() & 0x000F;  
	
			}

			// For each face, calc normal
			//fprintf(outfile,"\nvectors of face\n");
				
			for (unsigned int i=0; i < faceCount; i++)
 			{

				// Find Vector from point A to point B
				tempVect1.X = vertex[face[i].B].X - vertex[face[i].A].X ;
				tempVect1.Y = vertex[face[i].B].Y - vertex[face[i].A].Y ;
				tempVect1.Z = vertex[face[i].B].Z - vertex[face[i].A].Z ;
				
				// Find Vector from point A to point C
				tempVect2.X = vertex[face[i].C].X - vertex[face[i].A].X ;
				tempVect2.Y = vertex[face[i].C].Y - vertex[face[i].A].Y ;
				tempVect2.Z = vertex[face[i].C].Z - vertex[face[i].A].Z ;
				
				// Multiply vectors with a cross product
				faceVect[i].X = (tempVect1.Y * tempVect2.Z) - (tempVect1.Z * tempVect2.Y);
				faceVect[i].Y = (tempVect1.Z * tempVect2.X) - (tempVect1.X * tempVect2.Z);
				faceVect[i].Z = (tempVect1.X * tempVect2.Y) - (tempVect1.Y * tempVect2.X);

				// Normalize the result
				SetLength(i, 1.0);			

			}
			
			// For each vertex, count number of faces that reference it
						
			// The vertex to count
			for (unsigned int v=0; v < vertexCount; v++)
			{
				//cout << "\n\t\tLooking for vertex:" << v << "\n";
			
				// Go through the faces and assign them to a list to be averaged
				for (unsigned int f=0; f < faceCount; f++)
	 			{
					//cout << "\n\t\t\t\tface: " << f ;
		 			if ( face[f].A == v ){
		 				//cout << " *** found ref to vertex" << v << "  - ref " << refs;
		 				mixFace[refs].X = faceVect[f].X;
		 				mixFace[refs].Y = faceVect[f].Y;
		 				mixFace[refs].Z = faceVect[f].Z;
		 				refs++;}
		 			if ( face[f].B == v ){
		 				//cout << " *** found ref to vertex" << v << "  - ref " << refs;
		 				mixFace[refs].X = faceVect[f].X;
		 				mixFace[refs].Y = faceVect[f].Y;
		 				mixFace[refs].Z = faceVect[f].Z;
		 				refs++;}
		 			if ( face[f].C == v ){
		 				//cout << " *** found ref to vertex" << v << "  - ref " << refs;
		 				mixFace[refs].X = faceVect[f].X;
		 				mixFace[refs].Y = faceVect[f].Y;
		 				mixFace[refs].Z = faceVect[f].Z;
		 				refs++;}
	 			}
				
				//cout << "\n\t\t\tTotal refs to vertex "<<v<<" in faces "<<":" << refs << "\n";
				
				// Average the face vertexes 
					tempVect.X = 0;
					tempVect.Y = 0;
					tempVect.Z = 0;
			
				// Add all the face vectors
				for (unsigned int i = 0; i < refs; i++)
				{
					tempVect.X += mixFace[i].X;
					tempVect.Y += mixFace[i].Y;
					tempVect.Z += mixFace[i].Z;
				}			
				
				// Divide by number of references
				vertexVect[v].X = tempVect.X / refs;
				vertexVect[v].Y = tempVect.Y / refs;
				vertexVect[v].Z = tempVect.Z / refs;
			
				refs = 0;
				
			}

			for (unsigned int i=0; i < vertexCount; i++)
 			{
 
//				cout << "\t\t\t\tVertex " << i << ": "
				
//				<< " X: " << vertex[i].X
//				<< " Y: " << vertex[i].Y
//				<< " Z: " << vertex[i].Z << "\n";

				fprintf( outfile, "%f ",  vertex[i].X );
				fprintf( outfile, "%f ",  vertex[i].Y );
				fprintf( outfile, "%f ",  vertex[i].Z );

//				fprintf( outfile, "%f ",  faceVect[i].X );
//				fprintf( outfile, "%f ",  faceVect[i].Y );
//				fprintf( outfile, "%f\n", faceVect[i].Z );
				
				int z = 1;
				
				if (z == 0)
				{
				// Z Buffer Simulator
				fprintf( outfile, "%f ",  vertex[i].X/4 );
				fprintf( outfile, "%f ",  vertex[i].Y/4 );
				fprintf( outfile, "%f\n", vertex[i].Z/4 );
				}
				else
				{
				fprintf( outfile, "%f ",  vertexVect[i].X );
				fprintf( outfile, "%f ",  vertexVect[i].Y );
				fprintf( outfile, "%f\n", vertexVect[i].Z );
				}


			}





			fprintf(outfile,"\n%d\n", (faceCount * 2) );
			
			cout << "\n";

			for (unsigned int i = 0; i < faceCount; i++)
			{
			
//				cout << "\t\t\t\tFace " << i << ":  "
//
//				 << face[i].A	<< ", "
//				 << face[i].B	<< ", "
//				 << face[i].C   << "\n";

									
/*				// How it shows up in the 3DS file
				fprintf( outfile, "%d ",  face[i].A );
				fprintf( outfile, "%d ",  face[i].B );
				fprintf( outfile, "%d\n", face[i].C );
*/
				// Normals corrected
				fprintf( outfile, "%d ",  face[i].C );
				fprintf( outfile, "%d ",  face[i].B );
				fprintf( outfile, "%d\n", face[i].A );


				// for some reason, only displays properly when faces are doubled...
				fprintf( outfile, "%d ",  face[i].B );
				fprintf( outfile, "%d ",  face[i].A );
				fprintf( outfile, "%d\n", face[i].C );
/*
*/

			}
		break;


	//Unknown Chunk
		default:
			//cout << "\n*** Unknown Chunk ID: 0x" << hex << chunk_ID << dec 
			//<< " - size " << size << " bytes\n";
			//cout << "*** Found at file location: "<< current_pointer << "\n";

			//Move to the new chunk position
			fseek (infile, (current_pointer + size), SEEK_SET);
			
		break;
	}

}
	
	cout << "\n*** This code should never execute, buddy...";
	return(0);
}


//Value reading functions.

unsigned char ReadChar(void)
{
	return ( fgetc(infile) );
}

unsigned int ReadInt(void)
{
	unsigned int temp = ReadChar();
	return ( temp | (ReadChar() << 8) );
}

unsigned long ReadLong(void)
{
	unsigned long temp1, temp2, temp3, temp4;

	temp3 = temp1 = ReadInt();
	temp4 = temp2 = ReadInt();

	return ( temp3 + (temp4 * 0x10000L) );
}

int ReadLongName(void)
{

	tempName.SetTo("");
	
	unsigned int t = 0;
	char temp_name[] = "DefaultName";
	unsigned char letter;

	letter = ReadChar();
	
	if ( letter == 0)
		return (-1); // dummy object
	
	temp_name[t] = letter;
	t++;

	do
	{
		letter = ReadChar();
		temp_name[t] = letter;
		t++;
	}
	while( letter != 0);

	temp_name[t-1] = 0;

	tempName << temp_name;
	
	return (0);
}

int ReadTextureFile(void)
{

	TextureMapName.SetTo("");
	
	unsigned int t = 0;
	char temp_name[] = "DefaultName";
	unsigned char letter;

	letter = ReadChar();
	
	if ( letter == 0)
		return (-1); // dummy object
	
	temp_name[t] = letter;
	t++;

	do
	{
		letter = ReadChar();
		temp_name[t] = letter;
		t++;
	}
	while( letter != 0);

	temp_name[t-1] = 0;

	TextureMapName << temp_name;
	
	return (0);
}

float ReadFloat(void)
{
	float value;

	fread (&(value), sizeof (float), 1, infile);
	return (value);
}

double GetLength( unsigned int f)
{
	double dsq = (faceVect[f].X * faceVect[f].X) +
				 (faceVect[f].Y * faceVect[f].Y) + 
				 (faceVect[f].Z * faceVect[f].Z);
	if( dsq > 0.0 )
	{
		return sqrt(dsq);
	}
	else
	{
		return 0.0;
	}
}

bool SetLength( unsigned int f, double newLength)
{
	double oldLength = GetLength(f);
	if( oldLength < 1.0e-10 ) return false;

	double scale = newLength / oldLength;

	faceVect[f].X *= scale;
	faceVect[f].Y *= scale;
	faceVect[f].Z *= scale;

	return true;
}



int main ( int argc, char *argv[] )
{
	cout << "\n3DS to BeOS Teapot Demo Converter\nby Yasser Malaika\n";
		
	// Print usage if no parameters are given
 	if (argc < 2)
	{
		cerr << "\nUsage: 3ds2tea [.3DS file to process] " << endl << endl ;//<< argv[0] << " ";
		exit(1);
	}

	cout << "\nAttempting to open 3DS binary file: " << (argv [1]) << "  " ;

	// Test for file validity
	if ((infile = fopen( (argv [1]), "r")) == NULL)
	{
		cerr << "\n\tERROR: Unable to load file " << (argv[1]) << "\n";
		exit(2);
	}

	// Open output file...
	if ((outfile = fopen( "/boot/beos/etc/teapot.data", "w")) == NULL)
	{
		cerr << "\aERROR: Unable to create output file.\n";
		exit(3);
	}

	// Determine size of file in bytes
	fseek(infile, 0, SEEK_END);
	fileSize = ftell(infile);
	fseek(infile, 0, SEEK_SET);
	
	//cout << "Success! opened " <<  fileSize << " bytes\n\n";

	//cout << "Begin Reading Chunks: \n";
	
	// Main translation function
	int Success = Translate();
	
	if (Success == 1)
	{
		cerr << "\n*** Error: Unable to Translate file: " << (argv[1])  << "\n\n";
		exit(2);
	}
	
	// Close All files	
	//cout << "Closing all files...\n"; 
   
	if (fclose(infile) !=0)
	{
		cout << "\aERROR: Unable to close input file\n";
		return 0;
	}

	if (fclose(outfile) !=0)
	{
		cout << "\aERROR: Unable to close output file\n";
		return 0;
	}


	cout << "\nTranslation Complete.\n\nThe GLTeapot Demo should now run!  :)\n" << endl;
	//system ("glteapot_app &");
    system("/boot/beos/apps/GLTeapot &");
	return (0);
}

